config = {
    "mpi": "dispel4py.new.mpi_process",
    "multi": "dispel4py.new.multi_process",
    "simple": "dispel4py.new.simple_process",
    "storm": "dispel4py.storm.storm_submission",
    "spark": "dispel4py.new.spark_process"
}
